# from importlib import metadata

from .image import Image
from .collection import Collection
from . import interpolate

MODEL_NAME = 'SIMS'

# __version__ = metadata.version(__package__ or __name__)
